# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaserver_setup_ca
short_description: Setup CA
description: Setup CA
options:
  dm_password:
    description: Directory Manager password
    type: str
    required: yes
  password:
    description: Admin user kerberos password
    type: str
    required: yes
  master_password:
    description: kerberos master password (normally autogenerated)
    type: str
    required: yes
  ip_addresses:
    description: List of Master Server IP Addresses
    type: list
    elements: str
    required: no
  domain:
    description: Primary DNS domain of the IPA deployment
    type: str
    required: yes
  realm:
    description: Kerberos realm name of the IPA deployment
    type: str
    required: yes
  hostname:
    description: Fully qualified name of this host
    type: str
    required: no
  no_host_dns:
    description: Do not use DNS for hostname lookup during installation
    type: bool
    default: no
    required: no
  pki_config_override:
    description: Path to ini file with config overrides
    type: str
    required: no
  setup_adtrust:
    description: Configure AD trust capability
    type: bool
    default: no
    required: no
  setup_kra:
    description: Configure a dogtag KRA
    type: bool
    default: no
    required: no
  setup_dns:
    description: Configure bind with our zone
    type: bool
    default: no
    required: no
  setup_ca:
    description: Configure a dogtag CA
    type: bool
    default: no
    required: no
  idstart:
    description: The starting value for the IDs range (default random)
    type: int
    required: yes
  idmax:
    description: The max value for the IDs range (default idstart+199999)
    type: int
    required: yes
  no_hbac_allow:
    description: Don't install allow_all HBAC rule
    type: bool
    default: no
    required: no
  no_pkinit:
    description: Disable pkinit setup steps
    type: bool
    default: no
    required: no
  dirsrv_config_file:
    description:
      The path to LDIF file that will be used to modify configuration of
      dse.ldif during installation of the directory server instance
    type: str
    required: no
  dirsrv_cert_files:
    description:
      Files containing the Directory Server SSL certificate and private key
    type: list
    elements: str
    required: no
  _dirsrv_pkcs12_info:
    description: The installer _dirsrv_pkcs12_info setting
    type: list
    elements: str
    required: no
  external_ca:
    description: External ca setting
    type: bool
    default: no
    required: no
  external_ca_type:
    description: Type of the external CA
    type: str
    required: no
  external_ca_profile:
    description:
      Specify the certificate profile/template to use at the external CA
    type: str
    required: no
  external_cert_files:
    description:
      File containing the IPA CA certificate and the external CA certificate
      chain
    type: list
    elements: str
    required: no
  subject_base:
    description:
      The certificate subject base (default O=<realm-name>).
      RDNs are in LDAP order (most specific RDN first).
    type: str
    required: no
  _subject_base:
    description: The installer _subject_base setting
    type: str
    required: no
  ca_subject:
    description: The installer ca_subject setting
    type: str
    required: no
  _ca_subject:
    description: The installer _ca_subject setting
    type: str
    required: no
  ca_signing_algorithm:
    description: Signing algorithm of the IPA CA certificate
    type: str
    required: no
  _random_serial_numbers:
    description: The installer _random_serial_numbers setting
    type: bool
    required: yes
  reverse_zones:
    description: The reverse DNS zones to use
    type: list
    elements: str
    required: no
  no_reverse:
    description: Do not create new reverse DNS zone
    type: bool
    default: no
    required: no
  auto_forwarders:
    description: Use DNS forwarders configured in /etc/resolv.conf
    type: bool
    default: no
    required: no
  domainlevel:
    description: The domain level
    type: int
    required: no
  _http_ca_cert:
    description: The installer _http_ca_cert setting
    type: str
    required: no
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
'''

RETURN = '''
'''

import os

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_server import (
    check_imports, AnsibleModuleLog, setup_logging, options, sysrestore, paths,
    ansible_module_get_parsed_ip_addresses,
    api_Backend_ldap2, redirect_stdout, ca, installutils, ds_init_info,
    custodiainstance, write_cache, x509, decode_certificate
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            # basic
            dm_password=dict(required=True, type='str', no_log=True),
            password=dict(required=True, type='str', no_log=True),
            master_password=dict(required=True, type='str', no_log=True),
            ip_addresses=dict(required=False, type='list', elements='str',
                              default=[]),
            domain=dict(required=True, type='str'),
            realm=dict(required=True, type='str'),
            hostname=dict(required=False, type='str'),
            no_host_dns=dict(required=False, type='bool', default=False),
            pki_config_override=dict(required=False, type='str'),
            # server
            setup_adtrust=dict(required=False, type='bool', default=False),
            setup_kra=dict(required=False, type='bool', default=False),
            setup_dns=dict(required=False, type='bool', default=False),
            setup_ca=dict(required=False, type='bool', default=False),
            idstart=dict(required=True, type='int'),
            idmax=dict(required=True, type='int'),
            no_hbac_allow=dict(required=False, type='bool', default=False),
            no_pkinit=dict(required=False, type='bool', default=False),
            dirsrv_config_file=dict(required=False, type='str'),
            dirsrv_cert_files=dict(required=False, type='list',
                                   elements='str'),
            _dirsrv_pkcs12_info=dict(required=False, type='list',
                                     elements='str'),
            # certificate system
            external_ca=dict(required=False, type='bool', default=False),
            external_ca_type=dict(required=False, type='str'),
            external_ca_profile=dict(required=False, type='str'),
            external_cert_files=dict(required=False, type='list',
                                     elements='str', default=None),
            subject_base=dict(required=False, type='str'),
            _subject_base=dict(required=False, type='str'),
            ca_subject=dict(required=False, type='str'),
            _ca_subject=dict(required=False, type='str'),
            ca_signing_algorithm=dict(required=False, type='str'),
            _random_serial_numbers=dict(required=True, type='bool'),
            # dns
            reverse_zones=dict(required=False, type='list', elements='str',
                               default=[]),
            no_reverse=dict(required=False, type='bool', default=False),
            auto_forwarders=dict(required=False, type='bool', default=False),
            # additional
            domainlevel=dict(required=False, type='int'),
            _http_ca_cert=dict(required=False, type='str'),
        ),
    )

    ansible_module._ansible_debug = True
    check_imports(ansible_module)
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # set values ############################################################

    # basic
    options.dm_password = ansible_module.params.get('dm_password')
    options.admin_password = ansible_module.params.get('password')
    options.master_password = ansible_module.params.get('master_password')
    options.ip_addresses = ansible_module_get_parsed_ip_addresses(
        ansible_module)
    options.domain_name = ansible_module.params.get('domain')
    options.realm_name = ansible_module.params.get('realm')
    options.host_name = ansible_module.params.get('hostname')
    options.no_host_dns = ansible_module.params.get('no_host_dns')
    options.pki_config_override = ansible_module.params.get(
        'pki_config_override')
    # server
    options.setup_adtrust = ansible_module.params.get('setup_adtrust')
    options.setup_kra = ansible_module.params.get('setup_kra')
    options.setup_dns = ansible_module.params.get('setup_dns')
    options.setup_ca = ansible_module.params.get('setup_ca')
    options.idstart = ansible_module.params.get('idstart')
    options.idmax = ansible_module.params.get('idmax')
    options.no_hbac_allow = ansible_module.params.get('no_hbac_allow')
    options.no_pkinit = ansible_module.params.get('no_pkinit')
    options.dirsrv_config_file = ansible_module.params.get(
        'dirsrv_config_file')
    options.dirsrv_cert_files = ansible_module.params.get('dirsrv_cert_files')
    options._dirsrv_pkcs12_info = ansible_module.params.get(
        '_dirsrv_pkcs12_info')
    # certificate system
    options.external_ca = ansible_module.params.get('external_ca')
    options.external_ca_type = ansible_module.params.get('external_ca_type')
    options.external_ca_profile = ansible_module.params.get(
        'external_ca_profile')
    options.external_cert_files = ansible_module.params.get(
        'external_cert_files')
    options.subject_base = ansible_module.params.get('subject_base')
    options._subject_base = ansible_module.params.get('_subject_base')
    options.ca_subject = ansible_module.params.get('ca_subject')
    options._ca_subject = ansible_module.params.get('_ca_subject')
    options.ca_signing_algorithm = ansible_module.params.get(
        'ca_signing_algorithm')
    options._random_serial_numbers = ansible_module.params.get(
        '_random_serial_numbers')
    # dns
    options.reverse_zones = ansible_module.params.get('reverse_zones')
    options.no_reverse = ansible_module.params.get('no_reverse')
    options.auto_forwarders = ansible_module.params.get('auto_forwarders')
    # additional
    options.domainlevel = ansible_module.params.get('domainlevel')
    options._http_ca_cert = ansible_module.params.get('_http_ca_cert')
    if options._http_ca_cert:
        options._http_ca_cert = decode_certificate(options._http_ca_cert)

    # init #################################################################

    options.promote = False  # first master, no promotion

    # Repeat from ca.install_check
    # ca.external_cert_file and ca.external_ca_file need to be set
    if options.external_cert_files:
        ca.external_cert_file, ca.external_ca_file = \
            installutils.load_external_cert(
                options.external_cert_files, options._ca_subject)

    fstore = sysrestore.FileStore(paths.SYSRESTORE)

    api_Backend_ldap2(options.host_name, options.setup_ca, connect=True)

    ds = ds_init_info(ansible_log, fstore,
                      options.domainlevel, options.dirsrv_config_file,
                      options.realm_name, options.host_name,
                      options.domain_name, options.dm_password,
                      options.idstart, options.idmax,
                      options.subject_base, options.ca_subject,
                      options.no_hbac_allow, options._dirsrv_pkcs12_info,
                      options.no_pkinit)

    # setup CA ##############################################################

    if hasattr(custodiainstance, "get_custodia_instance"):
        if hasattr(custodiainstance.CustodiaModes, "FIRST_MASTER"):
            mode = custodiainstance.CustodiaModes.FIRST_MASTER
        else:
            mode = custodiainstance.CustodiaModes.MASTER_PEER
        custodia = custodiainstance.get_custodia_instance(options, mode)
        custodia.set_output(ansible_log)
        with redirect_stdout(ansible_log):
            custodia.create_instance()

    if options.setup_ca:
        if not options.external_cert_files and options.external_ca:
            # stage 1 of external CA installation
            cache_vars = {n: options.__dict__[n] for o, n in options.knobs()
                          if n in options.__dict__}
            write_cache(cache_vars)

        try:
            with redirect_stdout(ansible_log):
                if hasattr(custodiainstance, "get_custodia_instance"):
                    ca.install_step_0(False, None, options, custodia=custodia)
                else:
                    ca.install_step_0(False, None, options)
        except SystemExit:
            ansible_module.exit_json(changed=True,
                                     csr_generated=True)
    else:
        # Put the CA cert where other instances expect it
        x509.write_certificate(options._http_ca_cert, paths.IPA_CA_CRT)
        os.chmod(paths.IPA_CA_CRT, 0o444)

        if not options.no_pkinit:
            x509.write_certificate(options._http_ca_cert,
                                   paths.KDC_CA_BUNDLE_PEM)
        else:
            with open(paths.KDC_CA_BUNDLE_PEM, 'w'):
                pass
        os.chmod(paths.KDC_CA_BUNDLE_PEM, 0o444)

        x509.write_certificate(options._http_ca_cert, paths.CA_BUNDLE_PEM)
        os.chmod(paths.CA_BUNDLE_PEM, 0o444)

    with redirect_stdout(ansible_log):
        # we now need to enable ssl on the ds
        ds.enable_ssl()

    if options.setup_ca:
        with redirect_stdout(ansible_log):
            if hasattr(custodiainstance, "get_custodia_instance"):
                ca.install_step_1(False, None, options, custodia=custodia)
            else:
                ca.install_step_1(False, None, options)

    ansible_module.exit_json(changed=True,
                             csr_generated=False)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-replica-install code
#
# Copyright (C) 2018-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipareplica_setup_custodia
short_description: Setup custodia
description:
  Setup custodia
options:
  setup_ca:
    description: Configure a dogtag CA
    type: bool
    required: no
  setup_kra:
    description: Configure a dogtag KRA
    type: bool
    required: no
  no_pkinit:
    description: Disable pkinit setup steps
    type: bool
    required: no
  no_ui_redirect:
    description: Do not automatically redirect to the Web UI
    type: bool
    required: no
  subject_base:
    description:
      The certificate subject base (default O=<realm-name>).
      RDNs are in LDAP order (most specific RDN first).
    type: str
    required: yes
  config_master_host_name:
    description: The config master_host_name setting
    type: str
    required: yes
  ccache:
    description: The local ccache
    type: str
    required: yes
  _ca_enabled:
    description: The installer _ca_enabled setting
    type: bool
    required: no
  _ca_file:
    description: The installer _ca_file setting
    type: str
    required: no
  _kra_enabled:
    description: The installer _kra_enabled setting
    type: bool
    required: no
  _kra_host_name:
    description: The installer _kra_host_name setting
    type: str
    required: no
  _pkinit_pkcs12_info:
    description: The installer _pkinit_pkcs12_info setting
    type: list
    elements: str
    required: no
  _top_dir:
    description: The installer _top_dir setting
    type: str
    required: yes
  dirman_password:
    description: Directory Manager (master) password
    type: str
    required: yes
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
'''

RETURN = '''
'''

import os

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_replica import (
    check_imports, AnsibleModuleLog, setup_logging, installer, DN, paths,
    gen_env_boostrap_finalize_core, constants, api_bootstrap_finalize,
    gen_ReplicaConfig, gen_remote_api, api, redirect_stdout, custodiainstance
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            # server
            setup_ca=dict(required=False, type='bool'),
            setup_kra=dict(required=False, type='bool'),
            no_pkinit=dict(required=False, type='bool'),
            no_ui_redirect=dict(required=False, type='bool'),
            # certificate system
            subject_base=dict(required=True, type='str'),
            # additional
            config_master_host_name=dict(required=True, type='str'),
            ccache=dict(required=True, type='str'),
            _ca_enabled=dict(required=False, type='bool'),
            _ca_file=dict(required=False, type='str'),
            _kra_enabled=dict(required=False, type='bool'),
            _kra_host_name=dict(required=False, type='str'),
            _pkinit_pkcs12_info=dict(required=False, type='list',
                                     elements='str'),
            _top_dir=dict(required=True, type='str'),
            dirman_password=dict(required=True, type='str', no_log=True),
        ),
        supports_check_mode=False,
    )

    ansible_module._ansible_debug = True
    check_imports(ansible_module)
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # get parameters #

    options = installer
    # server
    options.setup_ca = ansible_module.params.get('setup_ca')
    options.setup_kra = ansible_module.params.get('setup_kra')
    options.no_pkinit = ansible_module.params.get('no_pkinit')
    # certificate system
    options.subject_base = ansible_module.params.get('subject_base')
    if options.subject_base is not None:
        options.subject_base = DN(options.subject_base)
    # additional
    master_host_name = ansible_module.params.get('config_master_host_name')
    ccache = ansible_module.params.get('ccache')
    os.environ['KRB5CCNAME'] = ccache
    # os.environ['KRB5CCNAME'] = ansible_module.params.get('installer_ccache')
    # installer._ccache = ansible_module.params.get('installer_ccache')
    ca_enabled = ansible_module.params.get('_ca_enabled')
    kra_enabled = ansible_module.params.get('_kra_enabled')
    kra_host_name = ansible_module.params.get('_kra_host_name')
    options._pkinit_pkcs12_info = ansible_module.params.get(
        '_pkinit_pkcs12_info')
    options._top_dir = ansible_module.params.get('_top_dir')
    dirman_password = ansible_module.params.get('dirman_password')

    # init #

    ansible_log.debug("== INSTALL ==")

    options = installer
    promote = installer.promote

    env = gen_env_boostrap_finalize_core(paths.ETC_IPA,
                                         constants.DEFAULT_CONFIG)
    api_bootstrap_finalize(env)
    config = gen_ReplicaConfig()
    config.dirman_password = dirman_password
    config.promote = installer.promote
    config.kra_enabled = kra_enabled
    config.kra_host_name = kra_host_name
    config.setup_ca = options.setup_ca

    remote_api = gen_remote_api(master_host_name, paths.ETC_IPA)

    conn = remote_api.Backend.ldap2
    ccache = os.environ['KRB5CCNAME']

    # There is a api.Backend.ldap2.connect call somewhere in ca, ds, dns or
    # ntpinstance
    api.Backend.ldap2.connect()
    conn.connect(ccache=ccache)

    with redirect_stdout(ansible_log):
        ansible_log.debug("-- INSTALL_CUSTODIA --")

        if not hasattr(custodiainstance, "get_custodia_instance"):
            custodia = custodiainstance.CustodiaInstance(config.host_name,
                                                         config.realm_name)
            if promote and \
               hasattr(custodiainstance.CustodiaInstance, "create_replica"):
                ansible_log.debug("-- CUSTODIA CREATE_REPLICA --")
                custodia.create_replica(config.master_host_name)
            else:
                ansible_log.debug("-- CUSTODIA CREATE_INSTANCE --")
                custodia.create_instance()
        else:
            if kra_enabled:
                # A KRA peer always provides a CA, too.
                mode = custodiainstance.CustodiaModes.KRA_PEER
            elif ca_enabled:
                mode = custodiainstance.CustodiaModes.CA_PEER
            else:
                mode = custodiainstance.CustodiaModes.MASTER_PEER
            custodia = custodiainstance.get_custodia_instance(config, mode)
            custodia.create_instance()

    # done #

    ansible_module.exit_json(changed=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Seth Kress <kresss@gmail.com>
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2020-2022 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}

DOCUMENTATION = """
---
module: ipapermission
short_description: Manage FreeIPA permission
description: Manage FreeIPA permission and permission members
extends_documentation_fragment:
  - ipamodule_base_docs
options:
  name:
    description: The permission name string.
    type: list
    elements: str
    required: true
    aliases: ["cn"]
  right:
    description: Rights to grant
    required: false
    choices: ["read", "search", "compare", "write", "add", "delete", "all"]
    type: list
    elements: str
    aliases: ["ipapermright"]
  attrs:
    description: All attributes to which the permission applies
    required: false
    type: list
    elements: str
  bindtype:
    description: Bind rule type
    required: false
    type: str
    choices: ["permission", "all", "anonymous", "self"]
    aliases: ["ipapermbindruletype"]
  subtree:
    description: Subtree to apply permissions to
    type: str
    required: false
    aliases: ["ipapermlocation"]
  extra_target_filter:
    description: Extra target filter
    required: false
    type: list
    elements: str
    aliases: ["filter", "extratargetfilter"]
  rawfilter:
    description: All target filters
    required: false
    type: list
    elements: str
    aliases: ["ipapermtargetfilter"]
  target:
    description: Optional DN to apply the permission to
    type: str
    required: false
    aliases: ["ipapermtarget"]
  targetto:
    description: Optional DN subtree where an entry can be moved to
    type: str
    required: false
    aliases: ["ipapermtargetto"]
  targetfrom:
    description: Optional DN subtree from where an entry can be moved
    type: str
    required: false
    aliases: ["ipapermtargetfrom"]
  memberof:
    description: Target members of a group (sets memberOf targetfilter)
    required: false
    type: list
    elements: str
  targetgroup:
    description: User group to apply permissions to (sets target)
    type: str
    required: false
    aliases: ["targetgroup"]
  object_type:
    description: Type of IPA object (sets subtree and objectClass targetfilter)
    type: str
    required: false
    aliases: ["type"]
  no_members:
    description: Suppress processing of membership
    required: false
    type: bool
  rename:
    description: Rename the permission object
    type: str
    required: false
    aliases: ["new_name"]
  action:
    description: Work on permission or member privilege level.
    type: str
    choices: ["permission", "member"]
    default: permission
    required: false
  state:
    description: The state to ensure.
    type: str
    choices: ["present", "absent", "renamed"]
    default: present
    required: false
author:
  - Seth Kress (@kresss)
  - Thomas Woerner (@t-woerner)
"""

EXAMPLES = """
# Ensure permission NAME is present
- ipapermission:
    name: manage-my-hostgroup
    right: all
    bindtype: permission
    object_type: host

# Ensure permission NAME is absent
- ipapermission:
    name: "Removed Permission Name"
    state: absent
"""

RETURN = """
"""


from ansible.module_utils.ansible_freeipa_module import \
    IPAAnsibleModule, compare_args_ipa


def find_permission(module, name):
    """Find if a permission with the given name already exist."""
    try:
        _result = module.ipa_command("permission_show", name, {"all": True})
    except Exception:  # pylint: disable=broad-except
        # An exception is raised if permission name is not found.
        return None
    else:
        return _result["result"]


def gen_args(right, attrs, bindtype, subtree,
             extra_target_filter, rawfilter, target,
             targetto, targetfrom, memberof, targetgroup,
             object_type, no_members, rename):
    _args = {}
    if right is not None:
        _args["ipapermright"] = right
    if attrs is not None:
        _args["attrs"] = attrs
    if bindtype is not None:
        _args["ipapermbindruletype"] = bindtype
    if subtree is not None:
        _args["ipapermlocation"] = subtree
    if extra_target_filter is not None:
        _args["extratargetfilter"] = extra_target_filter
    if rawfilter is not None:
        _args["ipapermtargetfilter"] = rawfilter
    if target is not None:
        _args["ipapermtarget"] = target
    if targetto is not None:
        _args["ipapermtargetto"] = targetto
    if targetfrom is not None:
        _args["ipapermtargetfrom"] = targetfrom
    if memberof is not None:
        _args["memberof"] = memberof
    if targetgroup is not None:
        _args["targetgroup"] = targetgroup
    if object_type is not None:
        _args["type"] = object_type
    if no_members is not None:
        _args["no_members"] = no_members
    if rename is not None:
        _args["rename"] = rename
    return _args


# pylint: disable=unused-argument
def result_handler(module, result, command, name, args, errors):
    # Get all errors
    # All "already a member" and "not a member" failures in the
    # result are ignored. All others are reported.
    for failed_item in result.get("failed", []):
        failed = result["failed"][failed_item]
        for member_type in failed:
            for member, failure in failed[member_type]:
                if "already a member" in failure \
                   or "not a member" in failure:
                    continue
                errors.append("%s: %s %s: %s" % (
                    command, member_type, member, failure))


def main():
    ansible_module = IPAAnsibleModule(
        argument_spec=dict(
            # general
            name=dict(type="list", elements="str", aliases=["cn"],
                      required=True),
            # present
            right=dict(type="list", elements="str", aliases=["ipapermright"],
                       default=None, required=False,
                       choices=["read", "search", "compare", "write", "add",
                                "delete", "all"]),
            attrs=dict(type="list", elements="str", default=None,
                       required=False),
            # Note: bindtype has a default of permission for Adds.
            bindtype=dict(type="str", aliases=["ipapermbindruletype"],
                          default=None, required=False, choices=["permission",
                          "all", "anonymous", "self"]),
            subtree=dict(type="str", aliases=["ipapermlocation"], default=None,
                         required=False),
            extra_target_filter=dict(type="list", elements="str",
                                     aliases=["filter", "extratargetfilter"],
                                     default=None, required=False),
            rawfilter=dict(type="list", elements="str",
                           aliases=["ipapermtargetfilter"],
                           default=None, required=False),
            target=dict(type="str", aliases=["ipapermtarget"], default=None,
                        required=False),
            targetto=dict(type="str", aliases=["ipapermtargetto"],
                          default=None, required=False),
            targetfrom=dict(type="str", aliases=["ipapermtargetfrom"],
                            default=None, required=False),
            memberof=dict(type="list", elements="str", default=None,
                          required=False),
            targetgroup=dict(type="str", default=None, required=False),
            object_type=dict(type="str", aliases=["type"], default=None,
                             required=False),
            no_members=dict(type="bool", default=None, required=False),
            rename=dict(type="str", default=None, required=False,
                        aliases=["new_name"]),
            action=dict(type="str", default="permission",
                        choices=["member", "permission"]),
            # state
            state=dict(type="str", default="present",
                       choices=["present", "absent", "renamed"]),
        ),
        supports_check_mode=True,
    )

    ansible_module._ansible_debug = True

    # Get parameters

    # general
    names = ansible_module.params_get("name")

    # present
    right = ansible_module.params_get("right")
    attrs = ansible_module.params_get("attrs")
    bindtype = ansible_module.params_get("bindtype")
    subtree = ansible_module.params_get("subtree")
    extra_target_filter = ansible_module.params_get("extra_target_filter")
    rawfilter = ansible_module.params_get("rawfilter")
    target = ansible_module.params_get("target")
    targetto = ansible_module.params_get("targetto")
    targetfrom = ansible_module.params_get("targetfrom")
    memberof = ansible_module.params_get("memberof")
    targetgroup = ansible_module.params_get("targetgroup")
    object_type = ansible_module.params_get("object_type")
    no_members = ansible_module.params_get("no_members")
    rename = ansible_module.params_get("rename")
    action = ansible_module.params_get("action")

    # state
    state = ansible_module.params_get("state")

    # Check parameters

    invalid = []

    if state == "present":
        if len(names) != 1:
            ansible_module.fail_json(
                msg="Only one permission can be added at a time.")
        if action == "member":
            invalid = ["bindtype", "target", "targetto", "targetfrom",
                       "subtree", "targetgroup", "object_type", "rename"]
        else:
            invalid = ["rename"]

    if state == "renamed":
        if len(names) != 1:
            ansible_module.fail_json(
                msg="Only one permission can be renamed at a time.")
        if action == "member":
            ansible_module.fail_json(
                msg="Member action can not be used with state 'renamed'")
        invalid = ["right", "attrs", "bindtype", "subtree",
                   "extra_target_filter", "rawfilter", "target", "targetto",
                   "targetfrom", "memberof", "targetgroup", "object_type",
                   "no_members"]

    if state == "absent":
        if len(names) < 1:
            ansible_module.fail_json(msg="No name given.")
        invalid = ["bindtype", "subtree", "target", "targetto",
                   "targetfrom", "targetgroup", "object_type",
                   "no_members", "rename"]
        if action != "member":
            invalid += ["right", "attrs", "memberof",
                        "extra_target_filter", "rawfilter"]

    ansible_module.params_fail_used_invalid(invalid, state, action)

    if bindtype == "self" and ansible_module.ipa_check_version("<", "4.8.7"):
        ansible_module.fail_json(
            msg="Bindtype 'self' is not supported by your IPA version.")

    if all([extra_target_filter, rawfilter]):
        ansible_module.fail_json(
            msg="Cannot specify target filter and extra target filter "
                "simultaneously.")

    # Init

    changed = False
    exit_args = {}

    # Connect to IPA API
    with ansible_module.ipa_connect():

        commands = []
        for name in names:
            # Make sure permission exists
            res_find = find_permission(ansible_module, name)

            # Create command
            if state == "present":
                # Generate args
                args = gen_args(right, attrs, bindtype, subtree,
                                extra_target_filter, rawfilter, target,
                                targetto, targetfrom, memberof, targetgroup,
                                object_type, no_members, rename)

                if action == "permission":
                    # Found the permission
                    if res_find is not None:
                        # For all settings is args, check if there are
                        # different settings in the find result.
                        # If yes: modify
                        if not compare_args_ipa(ansible_module, args,
                                                res_find):
                            commands.append([name, "permission_mod", args])
                    else:
                        commands.append([name, "permission_add", args])

                elif action == "member":
                    if res_find is None:
                        ansible_module.fail_json(
                            msg="No permission '%s'" % name)

                    member_attrs = {}
                    check_members = {
                        "attrs": attrs,
                        "memberof": memberof,
                        "ipapermright": right,
                        "ipapermtargetfilter": rawfilter,
                        "extratargetfilter": extra_target_filter,
                        # subtree member management is currently disabled.
                        # "ipapermlocation": subtree,
                    }

                    for _member, _member_change in check_members.items():
                        if _member_change is not None:
                            _res_list = res_find[_member]
                            # if running in a client context, data may be
                            # returned as a tuple instead of a list.
                            if isinstance(_res_list, tuple):
                                _res_list = list(_res_list)
                            _new_set = set(_res_list + _member_change)
                            if _new_set != set(_res_list):
                                member_attrs[_member] = list(_new_set)

                    if member_attrs:
                        commands.append([name, "permission_mod", member_attrs])

                else:
                    ansible_module.fail_json(
                        msg="Unknown action '%s'" % action)

            elif state == "renamed":
                if action == "permission":
                    # Generate args
                    # Note: Only valid arg for rename is rename.
                    args = gen_args(right, attrs, bindtype, subtree,
                                    extra_target_filter, rawfilter, target,
                                    targetto, targetfrom, memberof,
                                    targetgroup, object_type, no_members,
                                    rename)

                    # Found the permission
                    if res_find is not None:
                        # For all settings is args, check if there are
                        # different settings in the find result.
                        # If yes: modify
                        if not compare_args_ipa(ansible_module, args,
                                                res_find):
                            commands.append([name, "permission_mod", args])
                    else:
                        ansible_module.fail_json(
                            msg="Permission not found, cannot rename")
                else:
                    ansible_module.fail_json(
                        msg="Unknown action '%s'" % action)

            elif state == "absent":
                if action == "permission":
                    if res_find is not None:
                        commands.append([name, "permission_del", {}])

                elif action == "member":
                    if res_find is None:
                        ansible_module.fail_json(
                            msg="No permission '%s'" % name)

                    member_attrs = {}
                    check_members = {
                        "attrs": attrs,
                        "memberof": memberof,
                        "ipapermright": right,
                        "ipapermtargetfilter": rawfilter,
                        "extratargetfilter": extra_target_filter,
                        # subtree member management is currently disabled.
                        # "ipapermlocation": subtree,
                    }

                    for _member, _member_change in check_members.items():
                        if _member_change is not None:
                            _res_set = set(res_find[_member])
                            _new_set = _res_set - set(_member_change)
                            if _new_set != _res_set:
                                member_attrs[_member] = list(_new_set)

                    if member_attrs:
                        commands.append([name, "permission_mod", member_attrs])

            else:
                ansible_module.fail_json(msg="Unknown state '%s'" % state)

        # Execute commands

        changed = ansible_module.execute_ipa_commands(commands, result_handler)

    # Done

    ansible_module.exit_json(changed=changed, **exit_args)


if __name__ == "__main__":
    main()



import sys

param_docs = {
    "ccache": "The local ccache",
    "installer_ccache": "The installer ccache setting",
    "_top_dir": "The installer _top_dir setting",
    "_ca_enabled": "The installer _ca_enabled setting",
    "_add_to_ipaservers": "The installer _add_to_ipaservers setting",
    "_ca_subject": "The installer _ca_subject setting",
    "_subject_base": "The installer _subject_base setting",
    "config_setup_ca": "The config setup_ca setting",
    "config_master_host_name": "The config master_host_name setting",
    "config_ca_host_name": "The config ca_host_name setting",
    "config_ips": "The config ips setting",
    "_ca_file": "The installer _ca_file setting",
    "_kra_enabled": "The installer _kra_enabled setting",
    "_dirsrv_pkcs12_info": "The installer _dirsrv_pkcs12_info setting",
    "_pkinit_pkcs12_info": "The installer _pkinit_pkcs12_info setting",
    "_http_pkcs12_info": "The installer _http_pkcs12_info setting",
    "ds_ca_subject": "The ds.ca_subject setting",
    "ca_subject": "The installer ca_subject setting",
    "_hostname_overridden": "The installer _hostname_overridden setting",
    "_kra_host_name": "The installer _kra_host_name setting",
    "_http_ca_cert": "The installer _http_ca_cert setting",
    "_update_hosts_file": "The installer _update_host_file setting",
    "sssd": "The installer sssd setting",
    "dnsok": "The installer dnsok setting",

    "dm_password": "Directory Manager password",
    "password": "Admin user kerberos password",
    "ip_addresses": "List of Master Server IP Addresses",
    "domain": "Primary DNS domain of the IPA deployment",
    "realm": "Kerberos realm name of the IPA deployment",
    "hostname": "Fully qualified name of this host",
    "ca_cert_file": [
        "A CA certificate to use. Do not acquire the IPA CA certificate via",
        "automated means"
    ],
    "ca_cert_files": [
        "List of files containing CA certificates for the service certificate",
        "files"
    ],
    "no_host_dns": "Do not use DNS for hostname lookup during installation",
    "setup_adtrust": "Configure AD trust capability",
    "setup_ca": "Configure a dogtag CA",
    "setup_kra": "Configure a dogtag KRA",
    "setup_dns": "Configure bind with our zone",
    "force_join": "Force client enrollment even if already enrolled",
    "subject_base": [
        "The certificate subject base (default O=<realm-name>).",
        "RDNs are in LDAP order (most specific RDN first)."
    ],
    "server": "Fully qualified name of IPA server to enroll to",
    "dirman_password": "Directory Manager (master) password",
    "no_pkinit": "Disable pkinit setup steps",
    "no_ui_redirect": "Do not automatically redirect to the Web UI",
    "external_ca": "External ca setting",
    "external_cert_files": [
        "File containing the IPA CA certificate and the external CA "
        "certificate chain"
    ],
    "reverse_zones": "The reverse DNS zones to use",
    "no_reverse": "Do not create new reverse DNS zone",
    "auto_reverse": "Create necessary reverse zones",
    "forwarders": "Add DNS forwarders",
    "no_forwarders": "Do not add any DNS forwarders, use root servers instead",
    "auto_forwarders": "Use DNS forwarders configured in /etc/resolv.conf",
    "forward_policy": "DNS forwarding policy for global forwarders",
    "enable_compat": "Enable support for trusted domains for old clients",
    "netbios_name": "NetBIOS name of the IPA domain",
    "rid_base": "Start value for mapping UIDs and GIDs to RIDs",
    "secondary_rid_base": [
        "Start value of the secondary range for mapping UIDs and GIDs to RIDs"
    ],
    "pki_config_override": "Path to ini file with config overrides",
    "servers": "Fully qualified name of IPA servers to enroll to",
    "hidden_replica": "Install a hidden replica",
    "dirsrv_config_file": [
        "The path to LDIF file that will be used to modify configuration of",
        "dse.ldif during installation of the directory server instance"
    ],
    "dirsrv_cert_files": [
        "Files containing the Directory Server SSL certificate and private key"
    ],
    "http_cert_files": [
        "File containing the Apache Server SSL certificate and private key"
    ],
    "pkinit_cert_files": [
        "File containing the Kerberos KDC SSL certificate and private key"
    ],
    "no_ntp": "Do not configure ntp",
    "ntp_servers": "ntp servers to use",
    "ntp_pool": "ntp server pool to use",
    "no_dnssec_validation": "Disable DNSSEC validation",
    "master": "Master host name",
    "master_password": "kerberos master password (normally autogenerated)",
    "principal": [
        "User Principal allowed to promote replicas and join IPA realm"
    ],
    "dirsrv_pin": "The password to unlock the Directory Server private key",
    "http_pin": "The password to unlock the Apache Server private key",
    "pkinit_pin": "The password to unlock the Kerberos KDC private key",
    "dirsrv_cert_name":
        "Name of the Directory Server SSL certificate to install",
    "http_cert_name": "Name of the Apache Server SSL certificate to install",
    "pkinit_cert_name": "Name of the Kerberos KDC SSL certificate to install",
    "keytab": "Path to backed up keytab from previous enrollment",
    "mkhomedir": "Create home directories for users on their first login",
    "adtrust_netbios_name": "The adtrust netbios_name setting",
    "adtrust_reset_netbios_name": "The adtrust reset_netbios_name setting",
    "zonemgr":
        "DNS zone manager e-mail address. Defaults to hostmaster@DOMAIN",
    "ssh_trust_dns": "Configure OpenSSH client to trust DNS SSHFP records",
    "dns_ip_addresses": "The dns ip_addresses setting",
    "dns_reverse_zones": "The dns reverse_zones setting",
    "no_ssh": "Do not configure OpenSSH client",
    "no_sshd": "Do not configure OpenSSH server",
    "no_dns_sshfp": "Do not automatically create DNS SSHFP records",
    "allow_zone_overlap": "Create DNS zone even if it already exists",
    "skip_conncheck": "Skip connection check to remote master",
    "idstart": "The starting value for the IDs range (default random)",
    "idmax": "The max value for the IDs range (default idstart+199999)",
    "no_hbac_allow": "Don't install allow_all HBAC rule",
    "domainlevel": "The domain level",
    "external_ca_type": "Type of the external CA",
    "external_ca_profile": [
        "Specify the certificate profile/template to use at the external CA"
    ],
    "force": "Installer force parameter",
    "ca_signing_algorithm": "Signing algorithm of the IPA CA certificate",
    "debug": "Turn on extra debugging",
    "basedn": "The basedn of the IPA server (of the form dc=example,dc=com)",
    "allow_repair": [
        "Allow repair of already joined hosts. Contrary to "
        "ipaclient_force_join the host entry will not be changed on the server"
    ],
    "backup": "File to backup",
    "fqdn": [
        "The fully-qualified hostname of the host to add/modify/remove"
    ],
    "certificates": "A list of host certificates",
    "sshpubkey": "The SSH public key for the host",
    "ipaddress": "The IP address for the host",
    "random": "Generate a random password to be used in bulk enrollment",
    "state": "The desired host state",
    "kdc": "The name or address of the host running the KDC",
    "admin_keytab": "The path to a local admin keytab",
    "kinit_attempts": "Repeat the request for host Kerberos ticket X times",
    "automount_location": "The automount location",
    "firefox_dir": [
        "Specify directory where Firefox is installed (for example",
        "'/usr/lib/firefox')"
    ],
    "client_domain": "Primary DNS domain of the IPA deployment",
    "nisdomain": "The NIS domain name",
    "ca_enabled": "Whether the Certificate Authority is enabled or not",
    "on_master": "Whether the configuration is done on the master or not",
    "enable_dns_updates": [
        "Configures the machine to attempt dns updates when the ip address",
        "changes"
    ],
    "all_ip_addresses": [
        "All routable IP addresses configured on any interface will be added",
        "to DNS"
    ],
    "request_cert": "Request certificate for the machine",
    "preserve_sssd": "Preserve old SSSD configuration if possible",
    "no_sudo": "Do not configure SSSD as data source for sudo",
    "fixed_primary":
        "Configure sssd to use fixed server as primary IPA server",
    "permit": "Disable access rules by default, permit all access",
    "no_krb5_offline_passwords": [
        "Configure SSSD not to store user password when the server is offline"
    ],
    "force_ntpd": [
        "Stop and disable any time&date synchronization services besides ntpd",
        "Deprecated since 4.7"
    ],
    "no_nisdomain": "Do not configure NIS domain name",
    "configure_firefox": "Configure Firefox to use IPA domain credentials",
}


def gen_module_docs(module_in):
    with open(module_in) as in_f:
        in_lines = in_f.readlines()

    arg_spec = False
    args = []
    i = 0
    while i < len(in_lines):
        line = in_lines[i]
        stripped = line.strip()
        # print("stripped: %s" % repr(stripped))
        if stripped.startswith("# "):
            pass
        elif stripped.startswith("argument_spec=dict()"):
            pass
        elif stripped.startswith("argument_spec=dict("):
            arg_spec = True
        elif stripped.startswith("),") and arg_spec:
            arg_spec = False
        elif arg_spec:
            # if not "dict=(" in stripped:
            #     print("%s: Bad argument dict line '%s'" % (module_in,
            #                                                stripped))
            while ")," not in stripped and i < len(in_lines) - 1:
                next_stripped = in_lines[i + 1].strip()
                if not next_stripped.startswith("# "):
                    stripped += next_stripped
                i += 1
            # print("stripped: '%s'" % stripped)

            try:
                param, _dict = stripped.split("=", 1)
            except Exception:
                print("Failed to split line '%s'" % stripped)
                sys.exit(1)

            # print("_dict: '%s'" % _dict)
            if not _dict.startswith("dict(") or not _dict.endswith("),"):
                print("%s: Bad argument dict line 2 '%s'" % (module_in, _dict))
                sys.exit(1)
            else:
                _dict = _dict[5:-2]

            if param not in param_docs:
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                print("%s: param '%s' is not in param_docs" % (module_in,
                                                               param))
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                sys.exit(1)

            # print("param: '%s', dict: '%s'" % (param, _dict))

            opts = _dict.split(',')
            opts = [o.strip() for o in opts]
            required = False
            # no_log = False
            if "required=True" in opts:
                required = True
            # if "no_log=True" in opts:
            #     no_log = True

            # args.append([param, required, no_log])
            args.append([param, required])
        i += 1

    # print("%s: %s" % (module_in, repr(args)))

    def add_options(args):
        for param, required in args:
            out_lines.append("  %s:\n" % param)
            if isinstance(param_docs[param], list):
                out_lines.append("    description:\n")
                for x in param_docs[param]:
                    out_lines.append("      %s\n" % x)
            else:
                out_lines.append("    description: %s\n" % param_docs[param])
            out_lines.append("    required: %s\n" % ("yes", "no")[required])

    out_lines = []
    options = False
    in_options = False
    changed = False
    docs = False
    for line in in_lines:
        stripped = line.strip()
        if stripped.startswith("DOCUMENTATION = '''"):
            docs = True
        elif stripped.startswith("options:"):
            out_lines.append(line)
            add_options(args)
            options = True
            in_options = True
            changed = True
            continue
        elif stripped.startswith("author:"):
            if not options:
                add_options(args)
                options = True
                changed = True
            in_options = False
        elif stripped.startswith("'''"):
            if not options:
                add_options(args)
                options = True
                changed = True
            in_options = False
            docs = False
        elif docs and in_options:
            continue

        out_lines.append(line)

    print(module_in)
    # for line in out_lines:
    #     sys.stdout.write(line)

    if changed:
        with open(module_in, "w") as out_f:
            for line in out_lines:
                out_f.write(line)


gen_module_docs(sys.argv[1])



# -*- coding: utf-8 -*-

# Authors:
#   Chris Procter <cprocter@redhat.com>
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2020-2022 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}


DOCUMENTATION = '''
---
module: ipaconfig
author:
  - Chris Procter (@chr15p)
  - Thomas Woerner (@t-woerner)
short_description: Modify IPA global config options
description:
- Modify IPA global config options
extends_documentation_fragment:
  - ipamodule_base_docs
options:
    maxusername:
        description: Set the maximum username length between 1-255
        required: false
        type: int
        aliases: ['ipamaxusernamelength']
    maxhostname:
        description: Set the maximum hostname length between 64-255
        required: false
        type: int
        aliases: ['ipamaxhostnamelength']
    homedirectory:
        description: Set the default location of home directories
        required: false
        type: str
        aliases: ['ipahomesrootdir']
    defaultshell:
        description: Set the default shell for new users
        required: false
        type: str
        aliases: ['ipadefaultloginshell', 'loginshell']
    defaultgroup:
        description: Set the default group for new users
        required: false
        type: str
        aliases: ['ipadefaultprimarygroup']
    emaildomain:
        description: Set the default e-mail domain
        required: false
        type: str
        aliases: ['ipadefaultemaildomain']
    searchtimelimit:
        description:
        - Set maximum amount of time (seconds) for a search
        - values -1 to 2147483647 (-1 or 0 is unlimited)
        required: false
        type: int
        aliases: ['ipasearchtimelimit']
    searchrecordslimit:
        description:
        - Set maximum number of records to search
        - values -1 to 2147483647 (-1 or 0 is unlimited)
        required: false
        type: int
        aliases: ['ipasearchrecordslimit']
    usersearch:
        description:
        - Set comma-separated list of fields to search for user search
        required: false
        type: list
        elements: str
        aliases: ['ipausersearchfields']
    groupsearch:
        description:
        - Set comma-separated list of fields to search for group search
        required: false
        type: list
        elements: str
        aliases: ['ipagroupsearchfields']
    enable_migration:
        description: Enable migration mode
        type: bool
        required: false
        aliases: ['ipamigrationenabled']
    groupobjectclasses:
        description: Set default group objectclasses (comma-separated list)
        required: false
        type: list
        elements: str
        aliases: ['ipagroupobjectclasses']
    userobjectclasses:
        description: Set default user objectclasses (comma-separated list)
        required: false
        type: list
        elements: str
        aliases: ['ipauserobjectclasses']
    pwdexpnotify:
        description:
        - Set number of days's notice of impending password expiration
        - values 0 to 2147483647
        required: false
        type: int
        aliases: ['ipapwdexpadvnotify']
    configstring:
        description: Set extra hashes to generate in password plug-in
        required: false
        type: list
        elements: str
        choices:
        - "AllowNThash"
        - "KDC:Disable Last Success"
        - "KDC:Disable Lockout"
        - "KDC:Disable Default Preauth for SPNs"
        - ""
        aliases: ['ipaconfigstring']
    selinuxusermaporder:
        description: Set order in increasing priority of SELinux users
        required: false
        type: list
        elements: str
        aliases: ['ipaselinuxusermaporder']
    selinuxusermapdefault:
        description: Set default SELinux user when no match found in map rule
        required: false
        type: str
        aliases: ['ipaselinuxusermapdefault']
    pac_type:
        description: set default types of PAC supported for services
        required: false
        type: list
        elements: str
        choices: ["MS-PAC", "PAD", "nfs:NONE", ""]
        aliases: ["ipakrbauthzdata"]
    user_auth_type:
        description: set default types of supported user authentication
        required: false
        type: list
        elements: str
        choices: ["password", "radius", "otp", "disabled", ""]
        aliases: ["ipauserauthtype"]
    ca_renewal_master_server:
        description: Renewal master for IPA certificate authority.
        required: false
        type: str
    domain_resolution_order:
        description: set list of domains used for short name qualification
        required: false
        type: list
        elements: str
        aliases: ["ipadomainresolutionorder"]
    enable_sid:
        description: >
          New users and groups automatically get a SID assigned.
          Cannot be deactivated once activated. Requires IPA 4.9.8+.
        required: false
        type: bool
    netbios_name:
        description: >
          NetBIOS name of the IPA domain. Requires IPA 4.9.8+
          and SID generation to be activated.
        required: false
        type: str
    add_sids:
        description: >
          Add SIDs for existing users and groups. Requires IPA 4.9.8+
          and SID generation to be activated.
        required: false
        type: bool
'''

EXAMPLES = '''
---
- name: Playbook to handle global configuration options
  hosts: ipaserver
  become: true
  tasks:
    - name: return current values of the global configuration options
      ipaconfig:
        ipaadmin_password: SomeADMINpassword
      register: result
    - name: display default login shell
      debug:
        msg: '{{ result.config.defaultshell[0] }}'

    - name: set defaultshell and maxusername
      ipaconfig:
        ipaadmin_password: SomeADMINpassword
        defaultshell: /bin/bash
        maxusername: 64

- name: Playbook to enable SID and generate users and groups SIDs
  hosts: ipaserver
  tasks:
    - name: Enable SID and generate users and groups SIDS
      ipaconfig:
        ipaadmin_password: SomeADMINpassword
        enable_sid: yes
        add_sids: yes

- name: Playbook to change IPA domain netbios name
  hosts: ipaserver
  tasks:
    - name: Enable SID and generate users and groups SIDS
      ipaconfig:
        ipaadmin_password: SomeADMINpassword
        enable_sid: yes
        netbios_name: IPADOM
'''

RETURN = '''
config:
  description: Dict of all global config options
  returned: When no options are set
  type: dict
  contains:
    maxusername:
        description: maximum username length
        type: int
        returned: always
    maxhostname:
        description: maximum hostname length
        type: int
        returned: always
    homedirectory:
        description: default location of home directories
        type: str
        returned: always
    defaultshell:
        description: default shell for new users
        type: str
        returned: always
    defaultgroup:
        description: default group for new users
        type: str
        returned: always
    emaildomain:
        description: default e-mail domain
        type: str
        returned: always
    searchtimelimit:
        description: maximum amount of time (seconds) for a search
        type: int
        returned: always
    searchrecordslimit:
        description: maximum number of records to search
        type: int
        returned: always
    usersearch:
        description: list of fields to search in user search
        type: list
        elements: str
        returned: always
    groupsearch:
        description: list of fields to search in group search
        type: list
        elements: str
        returned: always
    enable_migration:
        description: Enable migration mode
        type: bool
        returned: always
    groupobjectclasses:
        description: default group objectclasses (comma-separated list)
        type: list
        elements: str
        returned: always
    userobjectclasses:
        description: default user objectclasses (comma-separated list)
        type: list
        elements: str
        returned: always
    pwdexpnotify:
        description: number of days's notice of impending password expiration
        type: str
        returned: always
    configstring:
        description: extra hashes to generate in password plug-in
        type: list
        elements: str
        returned: always
    selinuxusermaporder:
        description: order in increasing priority of SELinux users
        type: list
        elements: str
        returned: always
    selinuxusermapdefault:
        description: default SELinux user when no match is found in map rule
        type: str
        returned: always
    pac_type:
        description: default types of PAC supported for services
        type: list
        elements: str
        returned: always
    user_auth_type:
        description: default types of supported user authentication
        type: str
        returned: always
    ca_renewal_master_server:
        description: master for IPA certificate authority.
        type: str
        returned: always
    domain_resolution_order:
        description: list of domains used for short name qualification
        type: list
        elements: str
        returned: always
    enable_sid:
        description: >
          new users and groups automatically get a SID assigned.
          Requires IPA 4.9.8+.
        type: str
        returned: always
    netbios_name:
        description: NetBIOS name of the IPA domain. Requires IPA 4.9.8+.
        type: str
        returned: if enable_sid is True
'''


from ansible.module_utils.ansible_freeipa_module import \
    IPAAnsibleModule, compare_args_ipa, ipalib_errors


def config_show(module):
    _result = module.ipa_command_no_name("config_show", {"all": True})

    return _result["result"]


def get_netbios_name(module):
    try:
        _result = module.ipa_command_no_name("trustconfig_show", {"all": True})
    except Exception:  # pylint: disable=broad-except
        return None
    else:
        return _result["result"]["ipantflatname"][0]


def is_enable_sid(module):
    """When 'enable_sid' is true admin user and admins group have SID set."""
    _result = module.ipa_command("user_show", "admin", {"all": True})
    sid = _result["result"].get("ipantsecurityidentifier", [""])
    if not sid[0].endswith("-500"):
        return False
    _result = module.ipa_command("group_show", "admins", {"all": True})
    sid = _result["result"].get("ipantsecurityidentifier", [""])
    if not sid[0].endswith("-512"):
        return False
    return True


def main():
    ansible_module = IPAAnsibleModule(
        argument_spec=dict(
            # general
            maxusername=dict(type="int", required=False,
                             aliases=['ipamaxusernamelength']),
            maxhostname=dict(type="int", required=False,
                             aliases=['ipamaxhostnamelength']),
            homedirectory=dict(type="str", required=False,
                               aliases=['ipahomesrootdir']),
            defaultshell=dict(type="str", required=False,
                              aliases=['ipadefaultloginshell',
                                       'loginshell']),
            defaultgroup=dict(type="str", required=False,
                              aliases=['ipadefaultprimarygroup']),
            emaildomain=dict(type="str", required=False,
                             aliases=['ipadefaultemaildomain']),
            searchtimelimit=dict(type="int", required=False,
                                 aliases=['ipasearchtimelimit']),
            searchrecordslimit=dict(type="int", required=False,
                                    aliases=['ipasearchrecordslimit']),
            usersearch=dict(type="list", elements="str", required=False,
                            aliases=['ipausersearchfields']),
            groupsearch=dict(type="list", elements="str", required=False,
                             aliases=['ipagroupsearchfields']),
            enable_migration=dict(type="bool", required=False,
                                  aliases=['ipamigrationenabled']),
            groupobjectclasses=dict(type="list", elements="str",
                                    required=False,
                                    aliases=['ipagroupobjectclasses']),
            userobjectclasses=dict(type="list", elements="str", required=False,
                                   aliases=['ipauserobjectclasses']),
            pwdexpnotify=dict(type="int", required=False,
                              aliases=['ipapwdexpadvnotify']),
            configstring=dict(type="list", elements="str", required=False,
                              aliases=['ipaconfigstring'],
                              choices=["AllowNThash",
                                       "KDC:Disable Last Success",
                                       "KDC:Disable Lockout",
                                       "KDC:Disable Default Preauth for SPNs",
                                       ""]),  # noqa E128
            selinuxusermaporder=dict(type="list", elements="str",
                                     required=False,
                                     aliases=['ipaselinuxusermaporder']),
            selinuxusermapdefault=dict(type="str", required=False,
                                       aliases=['ipaselinuxusermapdefault']),
            pac_type=dict(type="list", elements="str", required=False,
                          aliases=["ipakrbauthzdata"],
                          choices=["MS-PAC", "PAD", "nfs:NONE", ""]),
            user_auth_type=dict(type="list", elements="str", required=False,
                                choices=["password", "radius", "otp",
                                         "disabled", ""],
                                aliases=["ipauserauthtype"]),
            ca_renewal_master_server=dict(type="str", required=False),
            domain_resolution_order=dict(type="list", elements="str",
                                         required=False,
                                         aliases=["ipadomainresolutionorder"]),
            enable_sid=dict(type="bool", required=False),
            add_sids=dict(type="bool", required=False),
            netbios_name=dict(type="str", required=False),
        ),
        supports_check_mode=True,
    )

    ansible_module._ansible_debug = True

    # Get parameters

    # general
    field_map = {
        "maxusername": "ipamaxusernamelength",
        "maxhostname": "ipamaxhostnamelength",
        "homedirectory": "ipahomesrootdir",
        "defaultshell": "ipadefaultloginshell",
        "defaultgroup": "ipadefaultprimarygroup",
        "emaildomain": "ipadefaultemaildomain",
        "searchtimelimit": "ipasearchtimelimit",
        "searchrecordslimit": "ipasearchrecordslimit",
        "usersearch": "ipausersearchfields",
        "groupsearch": "ipagroupsearchfields",
        "enable_migration": "ipamigrationenabled",
        "groupobjectclasses": "ipagroupobjectclasses",
        "userobjectclasses": "ipauserobjectclasses",
        "pwdexpnotify": "ipapwdexpadvnotify",
        "configstring": "ipaconfigstring",
        "selinuxusermaporder": "ipaselinuxusermaporder",
        "selinuxusermapdefault": "ipaselinuxusermapdefault",
        "pac_type": "ipakrbauthzdata",
        "user_auth_type": "ipauserauthtype",
        "ca_renewal_master_server": "ca_renewal_master_server",
        "domain_resolution_order": "ipadomainresolutionorder",
        "enable_sid": "enable_sid",
        "netbios_name": "netbios_name",
        "add_sids": "add_sids",
    }
    allow_empty_string = ["pac_type", "user_auth_type", "configstring"]
    reverse_field_map = {v: k for k, v in field_map.items()}

    params = {}
    for x in field_map:
        val = ansible_module.params_get(
            x, allow_empty_string=(x in allow_empty_string))

        if val is not None:
            params[field_map.get(x, x)] = val

    if params.get("ipamigrationenabled") is not None:
        params["ipamigrationenabled"] = \
            str(params["ipamigrationenabled"]).upper()

    if params.get("ipaselinuxusermaporder", None):
        params["ipaselinuxusermaporder"] = \
            "$".join(params["ipaselinuxusermaporder"])

    if params.get("ipadomainresolutionorder", None):
        params["ipadomainresolutionorder"] = \
            ":".join(params["ipadomainresolutionorder"])

    if params.get("ipausersearchfields", None):
        params["ipausersearchfields"] = \
            ",".join(params["ipausersearchfields"])

    if params.get("ipagroupsearchfields", None):
        params["ipagroupsearchfields"] = \
            ",".join(params["ipagroupsearchfields"])

    # verify limits on INT values.
    args_with_limits = [
        ("ipamaxusernamelength", 1, 255),
        ("ipamaxhostnamelength", 64, 255),
        ("ipasearchtimelimit", -1, 2147483647),
        ("ipasearchrecordslimit", -1, 2147483647),
        ("ipapwdexpadvnotify", 0, 2147483647),
    ]
    for arg, minimum, maximum in args_with_limits:
        if arg in params and (params[arg] > maximum or params[arg] < minimum):
            ansible_module.fail_json(
                msg="Argument '%s' must be between %d and %d."
                    % (arg, minimum, maximum))

    changed = False
    exit_args = {}

    # Connect to IPA API (enable_sid requires context == 'client')
    with ansible_module.ipa_connect(context="client"):
        has_enable_sid = ansible_module.ipa_command_param_exists(
            "config_mod", "enable_sid")

        result = config_show(ansible_module)

        if params:
            enable_sid = params.get("enable_sid")
            sid_is_enabled = has_enable_sid and is_enable_sid(ansible_module)

            if sid_is_enabled and enable_sid is False:
                ansible_module.fail_json(msg="SID cannot be disabled.")

            netbios_name = params.get("netbios_name")
            add_sids = params.get("add_sids")
            if has_enable_sid:
                if (
                    netbios_name
                    and netbios_name == get_netbios_name(ansible_module)
                ):
                    del params["netbios_name"]
                    netbios_name = None
                if not add_sids and "add_sids" in params:
                    del params["add_sids"]
                if any([netbios_name, add_sids]):
                    if sid_is_enabled:
                        params["enable_sid"] = True
                    else:
                        if not enable_sid:
                            ansible_module.fail_json(
                                msg="SID generation must be enabled for "
                                    "'netbios_name' and 'add_sids'. Use "
                                    "'enable_sid: yes'."
                            )
                else:
                    if sid_is_enabled and "enable_sid" in params:
                        del params["enable_sid"]

            else:
                if any([enable_sid, netbios_name, add_sids is not None]):
                    ansible_module.fail_json(
                        msg="This version of IPA does not support enable_sid, "
                            "add_sids or netbios_name setting through the "
                            "config module"
                    )
            params = {
                k: v for k, v in params.items()
                if k not in result or result[k] != v
            }
            # Remove empty string args from params if result arg is not set
            for k in ["ipakrbauthzdata", "ipauserauthtype", "ipaconfigstring"]:
                if k not in result and k in params and params[k] == [""]:
                    del params[k]
            if params \
               and not compare_args_ipa(ansible_module, params, result):
                changed = True
                if not ansible_module.check_mode:
                    try:
                        ansible_module.ipa_command_no_name("config_mod",
                                                           params)
                    except ipalib_errors.EmptyModlist:
                        changed = False
        else:
            del result['dn']
            type_map = {"str": str, "int": int, "list": list, "bool": bool}
            for key, value in result.items():
                k = reverse_field_map.get(key, key)
                if ansible_module.argument_spec.get(k):
                    arg_type = ansible_module.argument_spec[k]['type']
                    if k in (
                        'ipaselinuxusermaporder', 'domain_resolution_order'
                    ):
                        exit_args[k] = result.get(key)[0].split('$')
                    elif k in (
                        'usersearch', 'groupsearch'
                    ):
                        exit_args[k] = result.get(key)[0].split(',')
                    elif isinstance(value, str) and arg_type == "list":
                        exit_args[k] = [value]
                    elif (
                        isinstance(value, (tuple, list))
                        and arg_type in ("str", "int")
                    ):
                        exit_args[k] = type_map[arg_type](value[0])
                    elif (
                        isinstance(value, (tuple, list)) and arg_type == "bool"
                    ):
                        # FreeIPA 4.9.10+ and 4.10 use proper mapping for
                        # boolean values, so we need to convert it to str
                        # for comparison.
                        # See: https://github.com/freeipa/freeipa/pull/6294
                        exit_args[k] = (str(value[0]).upper() == "TRUE")
                    else:
                        if arg_type not in type_map:
                            raise ValueError(
                                "Unexpected attribute type: %s" % arg_type)
                        exit_args[k] = type_map[arg_type](value)
            # Add empty pac_type and user_auth_type if they are not set
            for key in ["pac_type", "user_auth_type"]:
                if key not in exit_args:
                    exit_args[key] = ""
            # Add empty domain_resolution_order if it is not set
            if "domain_resolution_order" not in exit_args:
                exit_args["domain_resolution_order"] = []
            # Set enable_sid
            if has_enable_sid:
                exit_args["enable_sid"] = is_enable_sid(ansible_module)
                exit_args["netbios_name"] = get_netbios_name(ansible_module)

    # Done
    ansible_module.exit_json(changed=changed, config=exit_args)


if __name__ == "__main__":
    main()
